/*--------------------------------------------------------------------
 *    The GMT-system:	@(#)gmt_mgg.c	2.18  03/09/99
 *
 *    Copyright (c) 1991-1999 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 * gmt_mgg.c is a collection of functions that are used by several of
 * the mgg GMT-SYSTEM programs. These functions perform:
 *	> Converting time from seconds to dates and back again
 *	> Find right path to a gmt-file
 *
 * Author:	Paul Wessel
 * Date:	13-JUL-1988
 * Updated:	13-JUN-1991 v2.0
 *		10-MAR-1999 v3.2
 *
 * List of functions:
 *
 *	gmt_date		- Convert seconds to gmt date.
 *	gmt_init		- Initialize gmt time functions.
 *	gmt_time		- Convert gmt date to seconds.
 *	gmtpath_init		- Get paths to gmtfile direcories.
 *	gmtpath_func		- Return full path to a leg.
 */
 
#include "gmt.h"	/* System dependent files */

struct GMT {
  int daymon[61];	/* Cumulative number of days up to last month */
  int first_year;	/* The year the cruise started */
};

char *gmt_path[10];	/* Max 10 directories for now */
int n_gmt_paths = 0;	/* Number of these directories */

/* GMT function gmt_date computes the date (hr/mi/sec/dd/mm/yy) based
 * on the total time in seconds since the beginning of first_year.
 * The pointer to the GMT structure is passed allong with the other
 * arguments. The Julian day is returned. the yymmddhhmmss is passed
 * through the argument list.
 */
 
int gmt_date(int time, int *year, int *month, int *day, int *hour, int *minute, int *second, struct GMT *gmt_struct)
{
	int day_time, julian_day;
	day_time = time/86400;
	*month = day_time / 31 + 1;	/* Only approximately, may be smaller */

	while (gmt_struct->daymon[*month +1] <= day_time) (*month)++;
	*year = (*month  - 1) / 12 + gmt_struct->first_year;
	*day = day_time - gmt_struct->daymon[*month] + 1;
	julian_day = (*month > 12) ?
		gmt_struct->daymon[*month] - gmt_struct->daymon[(*month - (*month)%12)] + *day :
		gmt_struct->daymon[*month] + *day;
	*month  = (*month-1)%12 + 1;
	time %= 86400;
	*hour = time / 3600;
	*minute = (time%3600) / 60;
	*second = time - *hour * 3600 - *minute * 60;
	return (julian_day);
}

/*
 *	GMT subroutine gmt_init sets up the structure GMT  which is
 *	used by other gmt routines (gmt_time,gmt_date) to convert
 *	times.  Daymon[month] contains the cumulative number of
 *	days from Jan 1 in first_year through the months PRIOR to the
 *	value of month.  0 <= month <= 60.  month = 0 only occurs
 *	during initializing in this routine. The user must declare
 *	a pointer to the struct GMT in the main program and pass it
 *	when calling the gmt_* functions. To define the GMT structure,
 *	include the file gmt.h
 *
 *	Paul Wessel
 *	12-JUL-1987
 *
 */
 
struct GMT *gmt_init(int year1)
{
	struct GMT *gmt_struct;
	int dm[12];	/* No of days in each month */
	int year, this_year, month, m;
	gmt_struct = (struct GMT *) GMT_memory (VNULL, (size_t)1, sizeof(struct GMT), "gmt_init");
	gmt_struct->first_year = year1;
	/* initialize days of the month etc. */
	dm[0] = 0;
	dm[1] = 31;
	dm[2] = 28;
	dm[3] = 31;
	dm[4] = 30;
	dm[5] = 31;
	dm[6] = 30;
	dm[7] = 31;
	dm[8] = 31;
	dm[9] = 30;
	dm[10] = 31;
	dm[11] = 30;
	gmt_struct->daymon[0] = 0;
	for (year = 0, month = 0; year < 5; year++) {
		this_year = gmt_struct->first_year + year;
		if (this_year%4 == 0 && !(this_year%400 == 0)) dm[2] = 29;
		for (m = 1; m <= 12; m++) {
			month++;
			gmt_struct->daymon[month] = gmt_struct->daymon[month - 1] + dm[m - 1];
		}
   		dm[2] = 28;
   		dm[0] = 31;
   	}
   	return (gmt_struct);
}

/* GMT function gmt_time returns the number of seconds from
 * first_year calculated from (hr/mi/sc/dd/mm/yy). The pointer
 * to the GMT structure is passed along with the arguments.
 */
 /* MODIFIED 10 July, 1987 by W. Smith  --  I killed a bug in
     month calculation */
 
int gmt_time(int *time, int year, int month, int day, int hour, int minute, int second, struct GMT *gmt_struct)
{
	int mon, n_days;
	if ((mon = (year - gmt_struct->first_year)) > 4) {
	  fprintf(stderr,"gmt_time:  Year - first_year > 4\n");
	  return(-1);
	}
	mon = mon * 12 + month;
	n_days = gmt_struct->daymon[mon] + day - 1;
	*time = n_days * 86400 + hour * 3600 + minute * 60 + second;
	return (*time);
}
	
/* gmtpath_init reads the LIBDIR/lib/mgg/gmtfile_paths file and gets all
 * the gmtfile directories.
 */
 
int gmtpath_init (void) {
	char file[BUFSIZ], line[BUFSIZ], *LIBDIR;
	FILE *fp;
	
	if ((LIBDIR = getenv ("GMTHOME")) == (char *)NULL) {
		fprintf (stderr, "gmtpath_init: Environment variable GMTHOME not set!\n");
		exit (-1);
	}

	sprintf (file, "%s%clib%cmgg%cgmtfile_paths\0", LIBDIR, DIR_DELIM, DIR_DELIM, DIR_DELIM);
	
	n_gmt_paths = 0;

	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "GMT Warning: path file %s for *.gmt files not found\n", file);
		fprintf (stderr, "(Will only look in current directory for such files)\n");
		return (0);
	}
	
	while (fgets (line, BUFSIZ, fp)) {
		if (line[0] == '#') continue;	/* Comments */
		if (line[0] == ' ' || line[0] == '\0') continue;	/* Blank line */
		gmt_path[n_gmt_paths] = GMT_memory (VNULL, (size_t)1, (size_t)(strlen (line)), "gmtpath_init");
		line[strlen (line)-1] = 0;
		strcpy (gmt_path[n_gmt_paths], line);
		n_gmt_paths++;
	}
	fclose (fp);
}
	
/* gmtpath takes a legid as argument and returns the full path
 * to where this data file can be found.  gmtpath_init must be
 * called first
 */
 
int gmtpath_func (char *leg_path, char *leg)
{
	int id;
	char geo_path[BUFSIZ];
	
	/* First look in current directory */
	
	sprintf (geo_path, "%s.gmt\0", leg);
	if (!access(geo_path, R_OK)) {
		strcpy(leg_path, geo_path);
		return (0);
	}
	
	/* Then look elsewhere */
	
	for (id = 0; id < n_gmt_paths; id++) {
		sprintf (geo_path, "%s%c%s.gmt\0", gmt_path[id], DIR_DELIM, leg);
		if (!access (geo_path, R_OK)) {
			strcpy (leg_path, geo_path);
			return (0);
		}
	}
	return(1);
}
	
