/*--------------------------------------------------------------------
 *    The GMT-system:	@(#)mgd77togmt.c	2.19  03/09/99
 *
 *    Copyright (c) 1991-1999 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 * mgd77togmt reads a mgd77 tape file from NGDC and creates
 * a gmt file directly.  If no time is provided, a fake time is
 * issued based on a year and a time increment. Use 1900 for
 * these cruises and estimate the time_inc based on the distance
 * between 2 nav points and an assumed velocity.
 *
 * Author:	Paul Wessel
 * Date:	9-SEP-1989
 *
 * Modified by WHFS 21 July 1993:
 *	Fixed usage message to be consistent with actual run for -H.
 *	Added carter table correction routines for depth.  Now uses
 *	twt if present, and then if not uses corr depth if present.
 * Modified by PW 19 February 1999:
 *	The 3 binary Carter tables have been replaced by a single ASCII
 *	table carter.d which makes it simpler to port this program.
 */
#include "gmt.h"
#include "gmt_mgg.h"

#define NODATA -32000

#define N_BINS 64800
#define N_CARTER_ZONES 85
#define N_CARTER_OFFSETS 86
#define N_CARTER_CORRECTIONS 5812
#define MAX_VALID_TWT   16000   /* Two-way time (msec) to 12000 nominal m  */

int     carter_setup(void);
int     get_bin();
int     get_carter_zone(int bin, int *zone);
int     get_depth_from_twt(int zone, short int twt_in_msec, short int *depth_in_corr_m);
int     get_twt_from_depth(int zone, short int depth_in_corr_m, short int *twt_in_msec);

int decode_record (char *string, int tflag, int *time, int *lon, int *lat, short int *faa, short int *mag, short int *top);
int     get_bin_6(int lat, int lon, int *bin); 

short int       carter_zone[N_BINS];
short int       carter_offset[N_CARTER_OFFSETS];
short int       carter_correction[N_CARTER_CORRECTIONS];

struct GMT *gmt;
struct GMT_REC *record;
int T_INC = 60, fake_running_time = 0;


int	carter_not_initialized = 1;

char *LIBDIR;

char *memory();
main (int argc, char **argv)
{
	int i, j, rec, n_records, n = 0, n_alloc = GMT_CHUNK, leg_year = 0, *year, use_list = FALSE;
	int t_flag = FALSE, error = FALSE, header = FALSE, n_recs;
	char file[80], *mfile, *list, agency[10], *legfname, line[BUFSIZ], **mgd77, **prefix;
	FILE *fp = NULL, *fpo;
	
	if ((LIBDIR = getenv ("GMTHOME")) == (char *)NULL) {
		fprintf (stderr, "mgd77togmt: Environment variable GMTHOME not set!\n");
		exit (-1);
	}

	legfname = list = mfile = 0;
	
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'Y':
					leg_year = atoi (&argv[i][2]);
					break;
				case 'I':	/* No time in data file, specify time increment between points */
					T_INC = atoi (&argv[i][2]);
					t_flag = TRUE;	/* fake time starts at 1/1/1900 @ 00:00:00 hr */
					break;
				case 'H':	/* File has 16 header records */
					header = TRUE;
					break;
				case 'F':
					legfname = &argv[i][2];
					break;
				case 'L':
					list = &argv[i][2];
					use_list = TRUE;
					break;
				case 'A':
					strcpy (agency, &argv[i][2]);
					break;
				case 'V':
					gmtdefs.verbose = TRUE;
					break;
				default:
					fprintf (stderr, "SYNTAX ERROR:  Unrecognized option -%c\n", argv[i][1]);
					error = TRUE;
					break;
			}
		}
		else
			mfile = argv[i];
	}
	
	if (use_list && !list) {
		fprintf (stderr, "SYNTAX ERROR -L option:  Must specify list file\n");
		error = TRUE;
	}
	if (use_list && (legfname || leg_year != 0)) {
		fprintf (stderr, "SYNTAX ERROR -L option:  Specify -L or the combination -F, -Y\n");
		error = TRUE;
	}
	
	if (argc == 1 || error) {
		fprintf (stderr, "usage: mgd77togmt [mgd77file] -F<filename> -Y<leg_year> OR -L<leglist> [-A<10 char agency name>] [NGDC-file -H -I<time_increment>] [-V]\n");
		fprintf (stderr, "	-F sets gmtfilename prefix (e.g., without path or .gmt)\n");
		fprintf (stderr, "	-Y sets start year\n");
		fprintf (stderr, "	-L gives name of a list with several records of <mgd77file> <gmtprefix> <leg_year>\n");
		fprintf (stderr,"\n\tOPTIONS:\n");
		fprintf(stderr,"	-A sets optional 10 char info for gmt header.\n");
		fprintf (stderr, "	-H means that the 16 MGD-77 header records are present\n");
		fprintf (stderr, "	-I sets fake timeincrement for files without time information\n");
		fprintf(stderr,"	-V runs in verbose mode.\n");
		
		exit (-1);
	}
	
	if (use_list) {
	
		mgd77 = (char **) GMT_memory (VNULL, (size_t)n_alloc, sizeof (char *), "mgd77togmt");
		prefix = (char **) GMT_memory (VNULL, (size_t)n_alloc, sizeof (char *), "mgd77togmt");
		year = (int *) GMT_memory (VNULL, (size_t)n_alloc, sizeof (int), "mgd77togmt");
		if ((fp = fopen (list, "r")) == NULL) {
			fprintf (stderr, "mgd77togmt: Cannot open file %s!\n", list);
			exit (-1);
		}
		while (fgets (line, BUFSIZ, fp)) {
			mgd77[n] = (char *) GMT_memory (VNULL, (size_t)80, sizeof (char), "mgd77togmt");
			prefix[n] = (char *) GMT_memory (VNULL, (size_t)80, sizeof (char), "mgd77togmt");
			sscanf (line, "%s %s %d", mgd77[n], prefix[n], &year[n]);
			n++;
			if (n == n_alloc) {
				n_alloc += GMT_CHUNK;
				mgd77 = (char **) GMT_memory ((void *)mgd77, (size_t)n_alloc, sizeof (char *), "mgd77togmt");
				prefix = (char **) GMT_memory ((void *)prefix, (size_t)n_alloc, sizeof (char *), "mgd77togmt");
				year = (int *) GMT_memory ((void *)year, (size_t)n_alloc, sizeof (int), "mgd77togmt");
			}
		}
		fclose (fp);
		mgd77 = (char **) GMT_memory ((void *)mgd77, (size_t)n, sizeof (char *), "mgd77togmt");
		prefix = (char **) GMT_memory ((void *)prefix, (size_t)n, sizeof (char *), "mgd77togmt");
		year = (int *) GMT_memory ((void *)year, (size_t)n, sizeof (int), "mgd77togmt");
	}
	else {
		mgd77 = (char **) GMT_memory (VNULL, (size_t)1, sizeof (char *), "mgd77togmt");
		prefix = (char **) GMT_memory (VNULL, (size_t)1, sizeof (char *), "mgd77togmt");
		year = (int *) GMT_memory (VNULL, (size_t)1, sizeof (int), "mgd77togmt");
		mgd77[0] = (char *) GMT_memory (VNULL, (size_t)80, sizeof (char), "mgd77togmt");
		prefix[0] = (char *) GMT_memory (VNULL, (size_t)80, sizeof (char), "mgd77togmt");
		strcpy (mgd77[0], mfile);
		strcpy (prefix[0], legfname);
		year[0] = leg_year;
		n = 1;
	}
	
	n_alloc = GMT_CHUNK;
	
	for (i = 0; i < n; i++) {
	
		if (gmtdefs.verbose) fprintf (stderr, "mgd77togmt: Processing file %s\n", mgd77[i]);
		
		if (mgd77[i][0] == 0)
			fp = stdin;
		else if ((fp = fopen (mgd77[i], "r")) == NULL) {
			fprintf (stderr, "mgd77togmt: Unable to open file %s - skipping\n", mgd77[i]);
			free ((void *)mgd77[i]);
			free ((void *)prefix[i]);
			continue;
		}
	
		gmt = gmt_init (year[i]);

		if (header) {  /* Skip header records */
			fgets (line, BUFSIZ, fp);
			n_recs = (strlen (line) == 121) ? 15 : 23;
			for (j = 0; j < n_recs; j++) fgets (line, BUFSIZ, fp);	/* Skip header records */
		}
	
		record = (struct GMT_REC *) GMT_memory (VNULL, (size_t)n_alloc, sizeof (struct GMT_REC), "mgd77togmt");

		rec = 0;
		while (fgets (line, BUFSIZ, fp)) {
			if (line[0] == '\n') continue;	/* Skip blank lines */
			if (!decode_record (line, t_flag,
				&record[rec].time,
				&record[rec].lon,
				&record[rec].lat,
				&record[rec].gmt[0],
				&record[rec].gmt[1],
				&record[rec].gmt[2]))
					rec++;
			else
				fprintf (stderr, "mgd77togmt: Trouble decoding time for record # %d\n", rec+1);
			if (rec == n_alloc) {
				n_alloc += GMT_CHUNK;
				record = (struct GMT_REC *) GMT_memory ((void *)record, (size_t)n_alloc, sizeof (struct GMT_REC), "mgd77togmt");
			}
		}
		if (fp != stdin) fclose (fp);
		free ((void *)gmt);
		n_records = rec;
	
		sprintf (file, "%s.gmt\0", prefix[i]);
		if ((fpo = fopen (file, "wb")) == NULL) {
			fprintf (stderr, "mgd77togmt: Could not create file %s\n", file);
			exit (-1);
		}
	
		if (fwrite((void *)(&year[i]), (size_t)4, (size_t)1, fpo) != 1) {
			fprintf(stderr,"mgd77togmt: Error while writing first year\n");
			exit(-1);
		}
		if (fwrite((void *)(&n_records), (size_t)4, (size_t)1, fpo) != 1) {
			fprintf(stderr,"mgd77togmt: Error while writing no of records\n");
			exit(-1);
		}
		if (fwrite((void *)agency, (size_t)10, (size_t)1, fpo) != 1) {

			fprintf(stderr,"mgd77togmt: Error while writing info-header\n");
			exit(-1);
		}
		for (rec = 0; rec < n_records; rec++) {
			if (fwrite((void *)(&record[rec]), (size_t)18, (size_t)1, fpo) != 1) {
				fprintf(stderr,"mgd77togmt: Error writing data record no %d\n",rec);
				exit(-1);
			}
		}
		fclose(fpo);
	
		free ((void *)record);
		
		free ((void *)mgd77[i]);
		free ((void *)prefix[i]);
	}
	
	free ((void *)mgd77);
	free ((void *)prefix);
	free ((void *)year);
	
	exit(0);
}

int decode_record (char *string, int tflag, int *time, int *lon, int *lat, short int *faa, short int *mag, short int *top)
{
	int year, month, day, hour, min, sec, l_mag, l_top, test, bin, zone, l_twt;
	short int	twt;
	double tz, fmin;
	char s_tz[5], s_yr[3], s_mo[3], s_dy[3], s_hr[3], s_mi[6], s_lat[8], s_lon[9];
	char s_faa[5], s_mag[6], s_top[7], s_top_twt[7];

	/* First decode Time Zone */
	
	strncpy (s_tz, &string[10], (size_t) 4);	s_tz[4] = 0;
	tz = atof (s_tz) * 0.01;
	if (string[9] == '-') tz = -tz;
	
	/* Then decode Time */
	
	if (tflag) {	/* No time provided, use fake time */
		*time  = fake_running_time;
		fake_running_time += T_INC;
	}
	else {
		strncpy (s_yr, &string[14], (size_t) 2);	s_yr[2] = 0;
		year = atoi (s_yr) + 1900;
	
		strncpy (s_mo, &string[16], (size_t) 2);	s_mo[2] = 0;
		month = atoi (s_mo);
	
		strncpy (s_dy, &string[18], (size_t) 2);	s_dy[2] = 0;
		day = atoi (s_dy);
	
		strncpy (s_hr, &string[20], (size_t) 2);	s_hr[2] = 0;
		hour = atoi (s_hr);
	
		strncpy (s_mi, &string[22], (size_t) 5);	s_mi[5] = 0;
		fmin = atof (s_mi) * 0.001;
		min = floor (fmin);
		sec = rint (60.0 * (fmin - min));
		test = gmt_time (time, year, month, day, hour, min, sec, gmt);
		if (test < 0) return (1);
		*time += rint (tz * 3600.0);
	}
	
	/* Get lat lon */
	
	strncpy (s_lat, &string[28], (size_t) 7);	s_lat[7] = 0;
	*lat = 10 * atoi (s_lat);
	if (string[27] == '-') *lat = -(*lat);
	
	strncpy (s_lon, &string[36], (size_t) 8);	s_lon[8] = 0;
	*lon = 10 * atoi (s_lon);
	if (string[35] == '-') *lon = -(*lon);
	if ((*lon) < 0) *lon += 360000000;
	
	/* Get gravity */
	
	strncpy (s_faa, &string[104], (size_t) 4);	s_faa[4] = 0;
	*faa = atoi (s_faa);
	if ((*faa) == 9999 || (s_faa[0] == ' ' && s_faa[1] == ' ' && s_faa[2] == ' ' && s_faa[3] == ' '))
		*faa = NODATA;
	else if (string[103] == '-')
		*faa = -(*faa);
		
	/* Get magnetics */
	
	strncpy (s_mag, &string[73], (size_t) 5);	s_mag[5] = 0;
	l_mag = atoi (s_mag);
	if (l_mag == 99999 || (s_mag[0] == ' ' && s_mag[1] == ' ' && s_mag[2] == ' ' && s_mag[3] == ' ' && s_mag[4] == ' '))
		l_mag = NODATA;
	else if (string[72] == '-')
		l_mag = -l_mag;
	if (l_mag != NODATA) l_mag = rint (0.1 * l_mag);
	*mag = l_mag;
	
	/* Get Bathymetry */
	strncpy(s_top_twt, &string[45], (size_t) 6);	s_top_twt[6] = '\0';
	l_twt = atoi(s_top_twt);
	if (l_twt == 999999 || (s_top_twt[0] == ' ' && s_top_twt[1] == ' ' && s_top_twt[2] == ' ' &&
		s_top_twt[3] == ' ' && s_top_twt[4] == ' '&& s_top_twt[5] == ' ')) {
			/* No two-way time was given.  See if there is a corrected depth:  */

	
	
		strncpy (s_top, &string[51], (size_t) 6);	s_top[6] = 0;
		l_top = atoi (s_top);
		if (l_top == 999999 || (s_top[0] == ' ' && s_top[1] == ' ' && s_top[2] == ' ' &&
			s_top[3] == ' ' && s_top[4] == ' '&& s_top[5] == ' '))
			l_top = NODATA;
		else
			l_top = -rint (0.1 * l_top);
		*top = l_top;
	}
	else {
		twt = rint(0.1 * l_twt);
		/* Convert the twt to depth  */
		if ((get_bin_6(*lat, *lon, &bin)) || (get_carter_zone(bin, &zone)) || (get_depth_from_twt(zone, twt, top)) ) {
			fprintf(stderr,"mgd77togmt:  ERROR in Carter correction system.\n");
			*top = NODATA;
		}
		else {
			*top = -(*top);
		}
	}
	return (0);
}

int	get_bin_6(int lat, int lon, int *bin)
{
	/* Given signed long ints in the 1.0e06 times decimal degree range, 
		-90000000 <= lat < 90000000, 0 <= lon < 360000000, set bin
		number.
		Returns 0 if OK, -1 if error.  */

	int	latdeg, londeg;

	if (lat < -90000000 || lat > 90000000) {
		fprintf(stderr, "getbin_6:  Latitude domain error.\n");
		return(-1);
	}
	if (lon < 0 || lon > 360000000) {
		fprintf(stderr, "getbin_6:  Longitude domain error.\n");
		return(-1);
	}
	latdeg = (lat + 90000000)/1000000;
	if (latdeg == 180) latdeg = 179;	/* Map north pole to previous row  */

	londeg = lon/1000000;
	*bin = 360 * latdeg + londeg;

	return(0);
}

int	carter_setup(void)
{
	/* This routine must be called once before using carter table stuff.
	It reads the carter.d file and loads the appropriate arrays.
	It sets carter_not_initialized = FALSE upon successful completion
	and returns 0.  If failure occurrs, it returns -1.  */

	FILE	*fp = NULL;
	char	buffer [BUFSIZ];
	int	i;

	carter_not_initialized = TRUE;

	/* Read the correction table:  */

	sprintf(buffer,"%s%clib%cmgg%ccarter.d", LIBDIR, DIR_DELIM, DIR_DELIM, DIR_DELIM);
	if ( (fp = fopen(buffer, "r")) == NULL) {
                fprintf(stderr,"carter_setup:  Cannot open r %s\n", buffer);
                return(-1);
        }

	for (i = 0; i < 4; i++) fgets (buffer, BUFSIZ, fp);	/* Skip 4 headers */
	fgets (buffer, BUFSIZ, fp);

	if ((i = atoi (buffer)) != N_CARTER_CORRECTIONS) {
		fprintf(stderr, "carter_setup:  Incorrect correction key (%d), should be %d\n", i, N_CARTER_CORRECTIONS);
                return(-1);
	}

        for (i = 0; i < N_CARTER_CORRECTIONS; i++) {
                if (!fgets (buffer, BUFSIZ, fp)) {
			fprintf(stderr, "carter_setup:  Could not read correction # %d\n", i);
			return (-1);
		}
                carter_correction[i] = atoi (buffer);
        }

	/* Read the offset table:  */

	fgets (buffer, BUFSIZ, fp);	/* Skip header */
	fgets (buffer, BUFSIZ, fp);

	if ((i = atoi (buffer)) != N_CARTER_OFFSETS) {
		fprintf(stderr, "carter_setup:  Incorrect offset key (%d), should be %d\n", i, N_CARTER_OFFSETS);
                return(-1);
	}

        for (i = 0; i < N_CARTER_OFFSETS; i++) {
                 if (!fgets (buffer, BUFSIZ, fp)) {
			fprintf(stderr, "carter_setup:  Could not read offset # %d\n", i);
			return (-1);
		}
                carter_offset[i] = atoi (buffer);
        }

	/* Read the zone table:  */

	fgets (buffer, BUFSIZ, fp);	/* Skip header */
	fgets (buffer, BUFSIZ, fp);

	if ((i = atoi (buffer)) != N_BINS) {
		fprintf(stderr, "carter_setup:  Incorrect zone key (%d), should be %d\n", i, N_BINS);
                return(-1);
	}

        for (i = 0; i < N_BINS; i++) {
                 if (!fgets (buffer, BUFSIZ, fp)) {
			fprintf(stderr, "carter_setup:  Could not read offset # %d\n", i);
			return (-1);
		}
                carter_zone[i] = atoi (buffer);
        }
        fclose(fp);

	/* Get here when all is well.  */

	carter_not_initialized = FALSE;
	return(0);
}


int	get_carter_zone(int bin, int *zone)
{
	/* Sets value pointed to by zone to the Carter zone corresponding to
		the bin "bin".  Returns 0 if successful, -1 if bin out of
		range.  */

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_carter_zone:  Initialization failure.\n");
		return(-1);
	}

	if (bin < 0 || bin >= N_BINS) {
		fprintf(stderr,"get_carter_zone:  bin out of range.\n");
		return(-1);
	}
	*zone = carter_zone[bin];
	return(0);
}


int	get_depth_from_twt(int zone, short int twt_in_msec, short int *depth_in_corr_m)
{
	/* Given two-way travel time of echosounder in milliseconds, and
		Carter Zone number, finds depth in Carter corrected meters.
		Returns (0) if OK, -1 if error condition.  */

	int	i, nominal_z1500, low_hundred, part_in_100;

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_depth_from_twt:  Initialization failure.\n");
		return(-1);
	}
	if (zone < 1 || zone > N_CARTER_ZONES) {
		fprintf(stderr,"get_depth_from_twt:  Zone out of range.\n");
		return(-1);
	}
	if (twt_in_msec < 0) {
		fprintf(stderr,"get_depth_from_twt:  Negative twt.\n");
		return(-1);
	}

	nominal_z1500 = irint(0.75 * twt_in_msec);

	if (nominal_z1500 <= 100) {
		/* There is no correction in water this shallow.  */
		*depth_in_corr_m = nominal_z1500;
		return(0);
	}

	low_hundred = nominal_z1500 / 100;
	i = carter_offset[zone - 1] + low_hundred - 1;	/* -1 'cause .f indices */
	
	if (i >= (carter_offset[zone] - 1) ) {
		fprintf(stderr, "get_depth_from_twt:  twt too big.\n");
		return(-1);
	}

	part_in_100 = nominal_z1500%100;

	if (part_in_100) {	/* We have to interpolate the table  */

		if ( i == (carter_offset[zone] - 2) ) {
			fprintf(stderr, "get_depth_from_twt:  twt too big.\n");
			return(-1);
		}

		*depth_in_corr_m = (short) irint (carter_correction[i] +
			0.01 * part_in_100 * (carter_correction[i+1]
				- carter_correction[i]) );
		return(0);
	}
	else {
		*depth_in_corr_m = carter_correction[i];
		return(0);
	}
}


int	get_twt_from_depth(int zone, short int depth_in_corr_m, short int *twt_in_msec)
{
	/*  Given Carter zone and depth in Carter corrected meters,
	finds the two-way travel time of the echosounder in milliseconds.
	Returns -1 upon error, 0 upon success.  */

	int	min, max, guess;
	double	fraction;

	if (carter_not_initialized && carter_setup() ) {
		fprintf(stderr,"get_twt_from_depth:  Initialization failure.\n");
		return(-1);
	}
	if (zone < 1 || zone > N_CARTER_ZONES) {
		fprintf(stderr,"get_twt_from_depth:  Zone out of range.\n");
		return(-1);
	}
	if (depth_in_corr_m < 0) {
		fprintf(stderr,"get_twt_from_depth:  Negative depth.\n");
		return(-1);
	}

	if (depth_in_corr_m <= 100) {	/* No correction applies.  */
		*twt_in_msec = irint(1.33333 * depth_in_corr_m);
		return(0);
	}

	max = carter_offset[zone] - 2;
	min = carter_offset[zone - 1] - 1;

	if (depth_in_corr_m > carter_correction[max]) {
		fprintf(stderr,"get_twt_from_depth:  Depth too big.\n");
		return(-1);
	}

	if (depth_in_corr_m == carter_correction[max]) {
		/* Hit last entry in table exactly  */
		*twt_in_msec = irint(133.333 * (max - min) );
		return(0);
	}


	guess = (depth_in_corr_m / 100) + min;
	if (guess > max) guess == max;
	while (guess < max && carter_correction[guess] < depth_in_corr_m) guess++;
	while (guess > min && carter_correction[guess] > depth_in_corr_m) guess--;

	if (depth_in_corr_m == carter_correction[guess]) {
		/* Hit a table value exactly  */
		*twt_in_msec = irint(133.333 * (guess - min) );
		return(0);
	}
	fraction = ( (double)(depth_in_corr_m - carter_correction[guess])
			/ (double)(carter_correction[guess + 1] - carter_correction[guess]) );
	*twt_in_msec = irint(133.333 * (guess - min + fraction) );
	return(0);
}

